from typing import Iterable, Self, Tuple, Hashable
from collections.abc import MutableSet, Set
from abc import ABC, abstractmethod, abstractproperty

class Symbol:
    """
    Ex:
    ===
        >>> A = Symbol("Some Name", 2)
        >>> print(A)
        Symb(Some Name, 2)
        >>> A(3, 4) # this is a fact

    """

    def __init__(self, name: str, arity: int):
        self.arity = arity
        self.name = name

    def __call__(self, *args) -> "Atom":
        assert len(args) == self.arity
        return Fact(self, *args)

    def __eq__(self, other):
        return type(self) == type(other) and self.name == other.name and self.arity == other.arity

    def __hash__(self):
        return hash((self.arity, self.name))

    def __repr__(self):
        return f"Symb({self.name}, {self.arity})"


class v(str):
    """
    Ex:
    ===
        >>> x = v("x")
        >>> x2 = v("x")
        >>> x == x2
        True
    """

    def __eq__(self, other):
        return type(self) is type(other) and super().__eq__(other)

    def __repr__(self):
        return f"v({self})"

    def __hash__(self):
        return super().__hash__()


class Atom:
    """
    Ex:
    ===
        >>> A = Symbol("Some Name", 3)
        >>> Atom(A, (3, 4, v("x")))
        A(3, 4, v("x"))
        
    """
    def __init__(self, symbol: Symbol, data: tuple): 
        self.symbol = symbol
        self.data = data

    def __hash__(self):
        return hash((self.symbol, self.data))


class Fact(Atom):
    def __init__(self, symbol: Symbol, data: tuple): 
        super().__init__(symbol, data)
        # TODO


class CQ:
    """
    Ex:
    ===
        # Q(x, x, y) := Exists z R(x, y, z)
        >>> R = Symbol("R", 3)
        >>> CQ((v("x"), v("x"), v("y")), (R(v("x"), v("y"), v("z")),))
    """
    def __init__(self, out_variables: tuple[v], *atoms: tuple[Atom, ...]):
        self.out_variables = out_variables
        self.atoms = atomes

    @property
    def variables(self) -> frozenset[v]:
        ... # TODO

    @property
    def is_boolean(self):
        ... # TODO
        
    def containment(self, query: "CQ") -> bool:
        ... # TODO


class UCQ(frozenset[CQ]): ...

class Database(ABC, Set):
    @abstractproperty
    def relations(self) -> tuple[Symbol]: 
        ...

    @abstractmethod
    def naive_evaluation(self, query: CQ | UCQ) -> frozenset[tuple[Fact]]:
        ...

    @abstractmethod
    def sat_evaluation(self, query) -> frozenset[tuple[Fact]]:
        ...

    @abstractproperty
    def active_domain(self) -> frozenset:
        ...

    def __repr__(self):
        return f"DB({','.join(map(str, self))})"

    @abstractproperty
    def relation(self, rel: Symbol) -> frozenset[tuple[Fact]]:
        """ Returns all the tuple associated to `rel` symbol"""

class StaticDatabase(Database, frozenset[Fact]):
    """ This class should implement database while being a frozenset.

        Ex:
        ===
            >>> A = Symbol("A", 2)
            >>> B = Symbol("B", 1)
            >>> DB = StaticDatabase([A(1, 2), B(2), A(3, 4)])
            >>> DB.relations

            >>> DB.relation(A)
            frozenset({A(1, 2), A(3, 4)})
            >>> DB.relation(B)
            frozenset({B(2)})
    """
    ...

class DynamicDatabase(Database, MutableSet[Fact]):
    """ This class is dynamic and fact can be update or delete 
        Try to implement a Database with the lower complexity
        when updating is performed !

        To implement MutableSet you need to implement the following function:

        * __contains__: check if a fact is in the database

        * __iter__: iter on all facts 

        * __len__: give the number of facts

        * add: add a new fact

        * discard: delete a fact if it is the database or do nothing otherwise

    """
    def __init__(self):
        self.datastore : set[fact] = set()  # We store the fact here !
